require 'opal-browser'
require 'dom/globals'
require 'dom/builder/binding'
require 'dom/browser_ext/node'
require 'dom/util'
require 'dom/debug'

module CSR
  module DOM
    module Builder
      include CSR::DOM::Globals
      include CSR::DOM::Util
      include CSR::DOM::Debug

      module_function

      def br
        tag(:br)
      end

      def div(attributes: nil, content: nil)
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        tag(:div, attributes: attributes, content: content)
      end

      # Any attribute or content which is a BoundProc will be
      # converted to bindings. See CSR::DOM::Binding.
      # Any attributes or content which is a Proc will be
      # resolved by proc.call.
      def tag(name, attributes: nil, content: nil)
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        debug 3, ->{[ __FILE__, __LINE__, __method__, "tag=#{name} attributes=#{attributes} content=#{content}" ]}
        attributes = attributes ? attributes : {}
        namespace = attributes[:namespace]
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        element = document.create_element(name, {namespace: namespace})
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        id = attributes.delete(:id) || attributes.delete('id')
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        if id
          debug 3, ->{[ __FILE__, __LINE__, __method__]}
          id = id.call if Proc === id
          debug 3, ->{[ __FILE__, __LINE__, __method__]}
          element.id = id
        end
        bindings = []
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        attributes.each do |key, value|
          debug 3, ->{[ __FILE__, __LINE__, __method__, "key=#{key} value=#{value}"]}
          if key == :style
            debug 3, ->{[ __FILE__, __LINE__, __method__]}
            value = resolve_value(element, value, bindings, id, :style, name)
            if value
              value = normalize_style(value)
              element.style(value)
            end
            debug 3, ->{[ __FILE__, __LINE__, __method__]}
          elsif key[0,2] == 'on'
            # event stuff is in browser/event/base.rb
            debug 3, ->{[ __FILE__, __LINE__, __method__]}
            event = key[2..-1]
            element.on(event, &value)
          else
            debug 3, ->{[ __FILE__, __LINE__, __method__]}
            value = resolve_value(element, value, bindings, id, key, name)
            if value
              key = sanitize_attribute(key)
              element[key] = value
            end
          end
        end
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        content = resolve_value(element, content, bindings, id, :content, name)
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        if content
          debug 3, ->{[ __FILE__, __LINE__, __method__]}
          content = sanitize_content(content)
          debug 3, ->{[ __FILE__, __LINE__, __method__]}
          element << content
        end
        debug 3, ->{[ __FILE__, __LINE__, __method__]}
        if bindings
          debug 3, ->{[ __FILE__, __LINE__, __method__]}
          element.bindings = bindings
        end
        debug 3, ->{[ __FILE__, __LINE__, __method__, "element=#{element}"]}
        element
      end

      def bind(proc=nil, &block)
        BoundProc.new(proc, &block)
      end

      def resolve_value(element, value, bindings, id, attr, tag_name)
        debug 2, ->{[ __FILE__, __LINE__, __method__, "tag=#{tag_name} value=#{value} bindings=#{bindings} id=#{id} attr=#{attr}" ]}
        result = case value
          when BoundProc
            # debug 1, ->{[ __FILE__, __LINE__, __method__, "BoundProc===value" ]}
            binding = Binding.new(
              # This proc was set in bind and should expect a Binding as argument
              # It can set up a watch, and should immediately set the binding value.
              value.proc,
              # this is local proc to update element with new value
              -> (value) {
                update_element(element, id, attr, value)
              }
            )
            bindings << binding
            # value proc should have set value in binding
            binding.value
          when Proc
            # debug 1, ->{[ __FILE__, __LINE__, __method__, "Proc===value" ]}
            value.call
          else
            # debug 1, ->{[ __FILE__, __LINE__, __method__, "value=#{value}" ]}
            value
        end
        debug 2, ->{[ __FILE__, __LINE__, __method__, "result=#{result}" ]}
        result
      end

      def sanitize_attribute(key)
        # key.to_s.gsub(/_([a-z])/) {|v| v[1].upcase}
        key.to_s.gsub(/_/, '-')
      end

      def sanitize_content(content)
        debug 3, ->{[ __FILE__, __LINE__, __method__, "content.class=#{content.class}" ]}
        case content
          when Browser::DOM::Element
            content
          when Enumerable
            content.map {|e| sanitize_content(e)}
          else
            content.to_s
        end
      end

      def update_element(element, id, attr, value)
        debug 2, ->{[ __FILE__, __LINE__, __method__, "element=#{element} id=#{id} attr=#{attr} value=#{value}" ]}
        case attr
          when :content
            content = sanitize_content(value)
            debug 2, ->{[ __FILE__, __LINE__, __method__, "content=#{content}"]}
            element.clear
            element << content
          when :style
            debug 2, ->{[ __FILE__, __LINE__, __method__ ]}
            element.style(normalize_style(value || {}))
          else
            debug 2, ->{[ __FILE__, __LINE__, __method__ ]}
            element[attr] = value
        end
      end

      def <<(*args, &block)
        debug 1, ->{[ __FILE__, __LINE__, __method__, "size=#{args.size} args=#{args}" ]}
        if block
          self << block.call
        else
          if args.size == 1
            dom_root << case args[0]
              when Hash
                name = args[0].delete(:tag)
                tag(name, args[0])
              else
                args[0]
            end
          end
        end
      end

      alias_method :add, :<<

      # Returns a Browser::DOM::Element with given id or nil.
      def element(id)
        document[id.to_s]
      end

      alias_method :el, :element
      alias_method :[], :element

      def remove(element)
        fail "#{self.class.name}##{__method__} not implemented"
      end

      def replace(element)
        fail "#{self.class.name}##{__method__} not implemented"
      end

      def traverse_dom(node_or_nodeset = nil, &block)
        node_or_nodeset ||= dom_root
        (NodeSet === node_or_nodeset ? node_or_nodeset : [node_or_nodeset]).each do |node|
          yield DOM(node)
          node.children.each do |child|
            traverse_dom(DOM(child), &block)
          end
        end
      end

      def stop_bindings(id = nil)
        switch_bindings(id, start: false)
      end

      def start_bindings(id = nil)
        switch_bindings(id, start: true)
      end

      def switch_bindings(id = nil, start: true)
        first = nil
        if id
          first = self[id]
          unless first
            fail "#{self.class.name}##{__method__}:#{__LINE_} : no element with id '#{id}'"
          end
        end
        traverse_dom(first) do |node|
          start ? node.start_bindings : node.stop_bindings
        end
      end

      def shutdown
        stop_bindings
        clear_interval
      end


    end
  end
end
