require 'dom/builder/base'
require 'dom/debug'

# adds helper to CSR::DOM::Builder

module CSR
  module DOM
    module Builder
      module_function
      include CSR::DOM::Debug

      def icon(name, callback: nil, attributes: nil, tooltip: nil, popover: nil)
        debug 2, ->{[ __FILE__, __LINE__, __method__, "name=#{name} callback=#{callback} attributes=#{attributes} tooltip=#{tooltip}" ]}
        style = {
          text_align: 'center',
          color: 'inherit',
          background_color: 'inherit',
          cursor: 'pointer',
        }
        klass = iconify(name)
        if attributes
          if attributes[:style]
            # arg style overrides any defaults
            style = style.merge(attributes[:style])
          end
          if attributes[:class]
            klass = klass + ' ' + attributes[:class]
          end
        end
        attributes = {
          class: klass,
          style: style
        }.merge(
          callback ? { onclick: callback } : {}
        )
        debug 2, ->{[ __FILE__, __LINE__, __method__, "attributes=#{attributes}" ]}
        icon = tag(
          :span,
          attributes: attributes
        )
        debug 2, ->{[ __FILE__, __LINE__, __method__, "icon=#{icon}" ]}
        if tooltip
          icon.tooltip(tooltip)
        elsif popover
          icon.popover(popover)
        end
        debug 2, ->{[ __FILE__, __LINE__, __method__, "icon=#{icon}" ]}
        icon
      end

      def icon_with_anchor(icon_name, href, icon_attributes: nil, anchor_attributes: nil, tooltip: nil, popover: nil)
        attributes = { href: href }
        if anchor_attributes
          attributes = merge_attributes(anchor_attributes, attributes)
        end
        tag(
          :a,
          attributes: attributes,
          content: icon(icon_name, attributes: icon_attributes, tooltip: tooltip, popover: popover)
        )
      end

      def plus_sign_icon_anchor(href, icon_attributes: nil, anchor_attributes: nil, tooltip: nil, popover: nil)
        _icon_attributes = {
          style: {
            color: 'lightgreen',
          }
        }
        if icon_attributes
          # argument takes precedence
          _icon_attributes = merge_attributes(_icon_attributes, icon_attributes)
        end
        icon_with_anchor(
          :plus_sign,
          href,
          icon_attributes: _icon_attributes,
          anchor_attributes: anchor_attributes,
          tooltip: tooltip,
          popover: popover
        )
      end

      def remove_sign_icon(callback: nil, attributes: nil, tooltip: nil, popover: nil)
        icon(
          :remove_sign,
          callback: callback,
          attributes: merge_attributes({ style: {color: 'red'} }, attributes),
          tooltip: tooltip,
          popover: popover
        )
      end

      def refresh_icon(callback: nil, attributes: nil, tooltip: nil, popover: nil)
        icon(
          :refresh,
          callback: callback,
          attributes: merge_attributes({ style: { color: 'inherit' } }, attributes),
          tooltip: tooltip,
          popover: popover
        )
      end

      def save_icon(callback: nil, attributes: nil, tooltip: nil, popover: nil)
        icon(
          :save,
          callback: callback,
          attributes: merge_attributes({ style: { color: 'inherit' } }, attributes),
          tooltip: tooltip,
          popover: popover
        )
      end

      def hamburger_icon(callback: nil, attributes: nil, tooltip: nil, popover: nil)
        icon(
          :menu_hamburger,
          callback: callback,
          attributes: merge_attributes({ style: { color: 'inherit' } }, attributes),
          tooltip: tooltip,
          popover: popover
        )
      end

      # Returns a span element with icon and dropdown menu
      #
      # 1. icon: is the name of the (bootstrap) icon
      # 2. items: should be hashes containing menu items
      #    e.g. { callback: ->{}, href: '#', content: 'list item'}
      # 3. menu_right: set true if right of menu should go under icon
      # 4. icon_attributes: any attributes for icon
      def drop_down_icon(icon: 'menu-hamburger', items: [], menu_right: false, icon_attributes: nil, menu_attributes: nil)
        _menu_attributes = {
            class: "dropdown-menu#{menu_right ? ' dropdown-menu-right' : nil}"
        }
        if menu_attributes
          _menu_attributes  = merge_attributes(_menu_attributes, menu_attributes)
        end
        div(
          attributes: {
            class: 'dropdown',
          },
          content: [
            div(
              attributes: {
                class: 'dropdown-toggle',
                data_toggle: 'dropdown',
              },
              content: icon(
                icon,
                attributes: icon_attributes
              )
            ),
            tag(
              :ul,
              attributes: _menu_attributes,
              content: items.map { |item|
                content = item[:content]
                if content == 'divider' || content == 'separator'
                  tag(
                    :li,
                    attributes: {
                      role: 'separator',
                      class: 'divider',
                    }
                  )
                else
                  attrs = {}
                  attrs[:href] = item[:href] if item[:href]
                  attrs[:onclick] = item[:callback] if item[:callback]
                  tag(
                    :li,
                    content: tag(
                      :div,
                      content: content,
                      attributes: attrs
                    )
                  )
                end
              }
            )
          ]
        )
      end

      def plus_sign_drop_down(items: [], menu_right: false, icon_attributes: nil, menu_attributes: nil)
        _icon_attributes = {
          style: {
            color: 'lightgreen'
          }
        }
        if icon_attributes
          # argument attributes take precedence
          debug 0, ->{[__FILE__, __LINE__, __method__, "_icon_attributes=#{_icon_attributes} icon_attributes=#{icon_attributes}"]}
          _icon_attributes = merge_attributes(_icon_attributes,  icon_attributes)
          debug 0, ->{[__FILE__, __LINE__, __method__, "_icon_attributes=#{_icon_attributes}"]}
        end
        drop_down_icon(
           icon: 'plus-sign',
           items: items,
           menu_right: menu_right,
           icon_attributes: _icon_attributes,
           menu_attributes: menu_attributes
        )
      end

      def plain_anchor(content, href)
        tag(
          :a,
          attributes: {
            href: href,
            style: { color: 'inherit', background_color: 'inherit'}
          },
          content: content
        )
      end

      def div_with_icon(callback: nil, icon: nil, pull: nil, attributes: nil, content: nil, icon_style: {}, tooltip: nil, popover: nil)
        icon ||= 'question-sign'
        pull ||= 'left'
        pull = pull.to_s
        icon = tag(
          :span,
          attributes: {
            onclick: callback,
            class: "glyphicon glyphicon-#{icon} pull-#{pull}",
            style: {
               font_size: 'smaller',
               margin_left: '0.5em',
               margin_right: '0.5em',
               margin_top: '0.3em',
               color: 'inherit',
               background_color: 'inherit',
            }.merge(
              icon_style # argument style overrides default
            )
          }
        )
        debug 1, ->{[__FILE__, __LINE__, __method__, "tooltip=#{tooltip} popover=#{popover}"]}
        if tooltip
          icon.tooltip(tooltip)
        elsif popover
          icon.popover(popover)
        end
        tag(
          :div,
          attributes: merge_attributes(
            { style: { cursor: 'pointer' } },
            attributes
          ),
          content: arrify(content, icon)
        )
      end

      def div_with_sort_icon(callback: nil, direction: 0, content: nil)
        if direction != 0
          tag(
            :div,
            attributes: {
              onclick: callback,
              style: { cursor: 'pointer' }
            },
            content: arrify(content) + [
              tag(:span,
                attributes: {
                  class: "glyphicon glyphicon-triangle-#{direction > 0 ? 'top' : 'bottom'}",
                  style: {
                     font_size: 'smaller',
                     margin_left: '0.5em',
                     vertical_align: 'middle',
                     color: 'inherit',
                     background_color: 'inherit',
                  }
                }
              )
            ]
          )
        else
          tag(
            :div,
            attributes: {
              onclick: callback,
              style: { cursor: 'pointer' }
            },
            content: content
          )
        end
      end

      def div_with_menu_up_down(callback: nil, up: true, down: false, content: nil, pull: 'left')
        div_with_up_down_icon(callback: callback, which: :menu, up: up, down: down, content: content, pull: pull)
      end

      def div_with_collapse_up_down(callback: nil, up: true, down: false, content: nil, pull: 'left')
        div_with_up_down_icon(callback: callback, which: :collapse, up: up, down: down, content: content, pull: pull)
      end

      # which can be :collapse or :menu (or string equivalents)
      def div_with_up_down_icon(callback: nil, which: :menu, up: true, down: false, content: nil, pull: 'left')
        up = up && !down
        pull = pull.to_s
        left = pull == 'left'
        icon_attributes = {
          class: "glyphicon glyphicon-#{which}-#{up ? 'up' : 'down'} pull-#{pull}",
          style: {
            font_size: 'smaller',
            margin_top: '0.2em',
            margin_left: left ? '0.3em' : '0.5em',
            margin_right: left ? '0.5em' : '0.3em',
            vertical_align: 'middle',
            color: 'inherit',
            background_color: 'inherit',
          }
        }
        icon = tag(
          :span,
          attributes: icon_attributes
        )
        div(
          attributes: { onclick: callback, style: { cursor: 'pointer' } },
          content: arrify(icon, content)
        )
      end

      # Returns a div element with given content and an icon to left or right.
      #
      # 1. icon: is the name of the (bootstrap) icon
      # 2. items: should be hashes containing menu items
      #    e.g. { callback: ->{}, href: '#', content: 'list item'}
      # 3. content: of the div (apart from the icon)
      # 4. pull: which side of div to pull the icon, 'right' or 'left'
      def div_with_dropdown_icon(icon: 'menu-hamburger', items: [], attributes: nil, content: nil, pull: 'right', menu_attributes: nil)
        tag(
          :div,
          attributes: attributes,
          content: arrify(
            tag(
              :span,
              attributes: {
                class: "pull-#{pull}",
                style: {
                  margin_left: '0.5em',
                  margin_right: '0.5em'
                }
              },
              content: drop_down_icon(
                icon: icon,
                items: items,
                menu_right: pull.to_s == 'right',
                menu_attributes: menu_attributes
              )
            ),
            content
          )
        )
      end

      # TODO: generalize from bootstrap
      def iconify(icon_name)
        "glyphicon glyphicon-#{icon_name.to_s.gsub('_', '-')}"
      end

    end

  end
end

