require 'dom/table/caption'
require 'dom/table/section'
require 'dom/table/column'
require 'dom/builder'
require 'dom/util'
require 'dom/debug'

module CSR
  module DOM
    class Table
      include CSR::DOM::Builder
      include CSR::DOM::Util
      include CSR::DOM::Debug

      # TODO: other than bootstrap, ...
      DEFAULT_CSS = 'table table-condensed table-bordered table-hover'
      DEFAULT_STYLE = {}

      # :id,                # dom id - optional
      # :css,               # table css class string or nil
      # :style,             # table style hash or nil
      # :caption_content,   # string or ComponentSpec with tag => 'caption'
      # :head_rows,         # Proc or array of objects or nil
      # :body_rows,         # Proc or array of objects or nil
      # :foot_rows,         # Proc or array of objects or nil
      # :sort_column_id,    # default sort column or nil
      # :sort_order,        # nil, 1 or -1
      # :accordion,         # whether table has master column which can be collapsed
      # :context,           # something meaningful to column content procs or nil
      # :reactive           # is the table reactive, default false

      attr_reader :id
      attr_reader :sort_column_id, :sort_column, :sort_orders, :sort_order
      attr_reader :accordion
      attr_reader :css
      attr_reader :style
      attr_reader :caption_content
      attr_reader :context
      attr_reader :section_ids

      def initialize(**options)
        # self.debug_level = 1
        debug 2, ->{[ __FILE__, __LINE__,  __method__, "options: #{options}" ]}
        @id = options[:id] || hex_id
        @rooted = false
        init_sections(options)
        @raw_columns = options[:columns]
        @context = options[:context]
        debug 2, ->{[ __FILE__, __LINE__,  __method__]}
        init_caption(options)
        debug 2, ->{[ __FILE__, __LINE__,  __method__]}
        init_css_style(options)
        debug 2, ->{[ __FILE__, __LINE__,  __method__]}
        init_row_sources(options)
        debug 2, ->{[ __FILE__, __LINE__,  __method__]}
        init_visibility
        debug 2, ->{[ __FILE__, __LINE__,  __method__]}
        init_sorting(options)
        debug 2, ->{[ __FILE__, __LINE__,  __method__]}
        init_accordion(options)
        debug 2, ->{[ __FILE__, __LINE__,  __method__]}
      end

      def new_column(*args)
        CSR::DOM::Table::Column.new(*args)
      end

      def new_content(*args)
        CSR::DOM::Content.new(*args)
      end

      def columns
        unless @columns
          @columns = if Proc === @raw_columns
            debug 2, ->{[ __FILE__, __LINE__,  __method__, "@raw_columns=#{@raw_columns}"]}
            c = @raw_columns.call
            debug 2, ->{[ __FILE__, __LINE__,  __method__, "@columns=#{@columns}"]}
            c
          else
            @raw_columns
          end
        end
        @columns
      end

      def rooted?
        @rooted
      end

      def root
        debug 2, ->{[ __FILE__, __LINE__, __method__ ]}
        @root ||= tag(
          :table,
          attributes: attributes,
          content: content
        )
        @rooted = true
        debug 2, ->{[ __FILE__, __LINE__, __method__, "@root=#{@root}" ]}
        @root
      end

      def update_root
        if @root
          @root.clear
          @root << content
        end
      end

      def attributes
        debug 2, ->{[ __FILE__, __LINE__, __method__ ]}
        result = {
          class: css,
          style: style
        }
        debug 2, ->{[ __FILE__, __LINE__, __method__, "result=#{result}" ]}
        result
      end

      def content
        debug 2, ->{[ __FILE__, __LINE__, __method__]}
        content = arrify(caption, sections)
        result = content.map {|e| e.root }
        debug 2, ->{[ __FILE__, __LINE__, __method__, "result=#{result}" ]}
        result
      end

      def caption
        debug 2, ->{[ __FILE__, __LINE__, __method__]}
        @caption ||= caption_content ? CSR::DOM::Table::Caption.new(self, caption_content) : nil
        debug 2, ->{[ __FILE__, __LINE__, __method__, "@caption=#{@caption}" ]}
        @caption
      end

      def sections
        @sections ||= section_ids.map do |id|
          debug 2, ->{[ __FILE__, __LINE__, __method__, "id=#{id}"]}
          section = CSR::DOM::Table::Section.new(self, id)
          debug 2, ->{[ __FILE__, __LINE__, __method__, "section=#{section}"]}
          section
        end.reject do |section|
          section.empty?
        end
        debug 2, ->{[ __FILE__, __LINE__, __method__, "@sections=#{@sections}"]}
        @sections
      end

      def section(id)
        sections.detect {|e| e.id == id}
      end

      def head
        section(:head)
      end

      def body
        section(:body)
      end

      def foot
        section(:foot)
      end

      def visible_columns
        @visible_columns ||= visible_column_indexes.map {|i| columns[i]}
      end

      def visible_column_indexes
        @visible_column_indexes
      end

      def column_ids
        columns.map(&:id)
      end

      def caption_attributes
        # TODO:
        nil
      end

      def cell_attributes(section_id)
        # TODO:
        nil
      end

      def section_attributes(section_id)
        nil
      end

      def sorted?
        !!@sort_column_id
      end

      # set or toggle the sort order to/of the given column
      def sort!(sort_column_id)
        update_sorting(sort_column_id)
        update_sections
      end

      def row_source(section_id)
        @row_sources[section_id]
      end

      def update
        debug 2, ->{[ __FILE__, __LINE__,  __method__ ]}
        init_sorting
        @caption = @sections = nil
        update_root
      end

      def update_caption
        @caption.update if @caption
      end

      def update_sections
        @sections = nil
        update_root
      end

      def update_section(id)
        debug 2, ->{[ __FILE__, __LINE__,  __method__, "id=#{id} " ]}
        section = section(id)
        debug 2, ->{[ __FILE__, __LINE__,  __method__, "section=#{section} " ]}
        if section
          section.update
        end
      end

      def update_head
        update_section :head
      end

      def update_body
        update_section :body
      end

      def update_foot
        update_section :foot
      end

      # source should be source (unsorted) index or row source (model)
      # columns may be visible column indexes or Column's
      def update_row(section_id, source, columns = nil)
        section = section(section_id)
        debug 2, ->{[ __FILE__, __LINE__, __method__, "section_id=#{section_id} source=#{source} columns=#{columns} section=#{section}" ]}
        if section
          section.update_row(source, columns)
        end
      end

      # source should be source (unsorted) index or row source (model)
      # column may be visible column index or a Column
      def update_cell(section_id, source, column)
        section = section(section_id)
        debug 2, ->{[ __FILE__, __LINE__, __method__, "section_id=#{section_id} source=#{source} column=#{column} section=#{section}" ]}
        if section
          section.update_cell(source, column)
        end
      end

      private

      def init_sections(options)
        # TODO: from options
        @section_ids = [:head, :body, :foot]
      end


      def init_row_sources(options)
        @row_sources = {}
        section_ids.each do |id|
          @row_sources[id] = options[:"#{id}_rows"]
        end
        debug 2, ->{[ __FILE__, __LINE__, __method__, "@row_sources=#{@row_sources}" ]}
      end

      def init_accordion(options)
        @accordion = !!options[:accordion]
      end

      def init_css_style(options)
        @css = options[:css] || DEFAULT_CSS
        @style = options[:style] || DEFAULT_STYLE
      end

      def init_caption(options)
        @caption_content = options[:caption]
      end

      def init_sorting(options = nil)
        # debug 2, ->{[ __FILE__, __LINE__,  __method__ ]}
        if options
          @initial_sort_column_id = options[:sort_column_id]
          @initial_sort_order = options[:sort_order]
        end
        @sort_column_id = sort_column_id unless
        @sort_column = @sort_column_id ? columns.detect {|c| c.id == @sort_column_id} : columns.detect {|c| c.sort?}
        @sort_column_id = @sort_column ? @sort_column.id : nil
        @sort_order = @initial_sort_order || 1 unless @sort_order
        @sort_orders = {}
        if @sort_column_id
          columns.each { |c| sort_orders[c.id] = c.id == sort_column_id ? @sort_order : 1 }
        end
        # debug 2, ->{[ __FILE__, __LINE__, __method__, "@sort_column_id=#{@sort_column_id}" ]}
      end

      def update_sorting(sort_column_id)
        # debug 2, ->{[ __FILE__, __LINE__, __method__, "column_id=#{column_id}" ]}
        if @sort_column_id == sort_column_id
          sort_orders[@sort_column_id] *= -1
        else
          @sort_column_id = sort_column_id
          @sort_column = columns.detect {|c| c.id == @sort_column_id}
        end
        @sort_order = sort_orders[@sort_column_id]
      end

      def init_visibility
        @visible_column_indexes = columns.size.times.to_a
      end

    end

  end # module DOM
end # module CSR

