require 'dom/table/row'
require 'dom/builder'
require 'dom/util'
require 'dom/debug'

module CSR
  module DOM
    class Table
      class Section
        include CSR::DOM::Builder
        include CSR::DOM::Util
        include CSR::DOM::Debug

        attr_reader :table, :id

        def initialize(table, id)
          # self.debug_level = 2
          @table, @id = table, id
        end

        def head?
          @id == :head
        end

        def foot?
          @id == :foot
        end

        def body?
          @id == :body
        end

        def update
          if @root
            debug 2, ->{[ __FILE__, __LINE__,  __method__, "@id=#{@id} @root=#{@root}" ]}
            animate do
              clear_rows
              update_root
            end
          else
            debug 2, ->{[ __FILE__, __LINE__,  __method__, "@id=#{@id} @root=nil" ]}
          end
        end

        def update_root
          @root.clear
          @root << content
        end

        def row_attributes(index)
          # TODO: implement
        end

        # source should be source (unsorted) index or row source (model)
        # columns may be column indexes or Column's
        def update_row(source, _columns = nil)
          debug 2, ->{[ __FILE__, __LINE__, __method__, "source=#{source} _columns=#{_columns} @rows.size=#{@rows ? @rows.size : 'nil'}" ]}
          if @rows
            columns = if _columns
              _columns.map { |c| Integer === c ? @table.columns[c] : c }
            end
            debug 2, ->{[ __FILE__, __LINE__, __method__, "source=#{source} columns=#{columns}" ]}
            if @id == :body && (columns.nil? || (@table.sort_column && columns.include?(@table.sort_column)))
              debug 2, ->{[ __FILE__, __LINE__, __method__, "updating whole section" ]}
              update
            else
              row_index = Integer === source ? source : source_index(source)
              debug 2, ->{[ __FILE__, __LINE__, __method__, "calling @rows[#{row_index}].update(#{columns})" ]}
              @rows[row_index].update(columns)
            end
          end
        end

        # source should be source (unsorted) index or row source (model)
        # column may be visible column index or a Column
        def update_cell(source, column)
          update_row(source, [column])
        end

        def root
          @root ||= tag(
            "t#{@id}",
            attributes: attributes,
            content: content
          )
          @root
        end

        def empty?
          row_sources.size == 0
        end

        def attributes
          debug 2, ->{[ __FILE__, __LINE__,  __method__, "@id=#{@id}" ]}
          result = table.section_attributes(@id)
          debug 2, ->{[ __FILE__, __LINE__,  __method__, "@id=#{@id} result=#{result}" ]}
          result
        end

        def content
          debug 2, ->{[ __FILE__, __LINE__,  __method__, "@id=#{@id}" ]}
          result = rows.map { |row| row.root }
          debug 2, ->{[ __FILE__, __LINE__,  __method__, "@id=#{@id} result=#{result}" ]}
          result
        end

        # Returns array of Table::Row's (in table sort order)
        def rows
          unless @rows
            # debug 2, ->{[ __FILE__, __LINE__,  __method__, "@id=#{@id}" ]}
            @rows = []
            row_indexes.each do |model_index|
              @rows << Row.new(@table, self, model_index)
            end
              # debug 2, ->{[ __FILE__, __LINE__,  __method__, "@id=#{@id} @rows=#{@rows.map(&:index)}}" ]}
          end
          @rows
        end

        def row_source(index)
          row_sources[index]
        end

        def source_index(source)
          row_sources.index(source)
        end

        def row_sources
          unless @row_sources
            debug 2, ->{[ __FILE__, __LINE__, __method__, "@id=#{@id}" ]}
            source = @table.row_source(@id)
            source = source.call if Proc === source
            debug 2, ->{[ __FILE__, __LINE__, __method__, "source=#{source}" ]}
            @row_sources = case source
              when Fixnum
                source.times.to_a
              when NilClass
                []
              else
                source
            end
            # if body?
            #   debug 2, ->{[ __FILE__, __LINE__, __method__, "@id=#{@id} @row_sources=#{@row_sources.map(&:code)}" ]}
            # end
          end
          @row_sources
        end

        # Returns array of indexes to row models
        # in sorted display order.
        def row_indexes
          unless @row_indexes
            models = row_sources
            @row_indexes = models.size.times.to_a
            if @id == :body && @table.sorted?
              order = @table.sort_order
              sort_column = @table.sort_column
              @row_indexes = @row_indexes.sort do |a, b|
                val_a = sort_value(sort_column, models[a])
                val_b = sort_value(sort_column, models[b])
                # debug 2, ->{[__FILE__, __LINE__, __method__, "a=#{val_a} b=#{val_b} order=#{order}"]}
                compare(val_a, val_b) * order
              end
            end
          end
          @row_indexes
        end

        # Returns the display index for an unsorted row index or row model
        def row_display_index(row)
          sources = row_sources
          if Integer === row
            return row_index.index(row)
          else
            row_indexes.each_with_index do |source_index, display_index|
              if row == sources[source_index]
                return display_index
              end
            end
          end
          nil
        end

        private

        def clear_rows
          @row_sources = @row_indexes = @rows = nil
        end

        def sort_value(column, row_source)
          content = column.section_content(@id)
          result = if Content === content
            content.sort_value(context: row_source)
          else
            content
          end
          result
        end

        def compare(a, b)
          if a && b
            a <=> b
          elsif a
            1
          elsif b
            -1
          else
            0
          end
        end
      end
    end
  end
end